"""Contains the dataclasses used in the YouTubeNotifier."""

from dataclasses import dataclass

from fastapi import FastAPI


@dataclass
class YouTubeNotifierConfig:
    """Represents the configuration of the YouTubeNotifier."""

    callback_url: str | None
    """The URL to receive notifications from YouTube"""

    host: str
    """The host to run the FastAPI app on"""

    port: int
    """The port to run the FastAPI app on"""

    app: FastAPI
    """The FastAPI app to receive notifications from YouTube"""

    using_ngrok: bool
    """Whether to use ngrok to receive notifications from YouTube"""

    password: str
    """The password to authenticate the YouTubeNotifier"""



"""Contains fixtures and utility functions."""

import os
from datetime import UTC, datetime

from dotenv import load_dotenv
from pyngrok import ngrok

from ytnoti import Channel, Timestamp, Video

CALLBACK_URL = "http://localhost:8000"

load_dotenv()
ngrok.set_auth_token(os.environ["NGROK_TOKEN"])


def get_channel() -> Channel:
    """Create a mock channel."""
    return Channel(
        id="mock_channel_id",
        name="Mock Channel",
        url="https://www.youtube.com/channel/mock_channel",
    )


def get_video() -> Video:
    """Create a mock video."""
    return Video(
        id="mock_video_id",
        title="Mock Video",
        url="https://www.youtube.com/watch?v=mock_video",
        timestamp=Timestamp(
            published=datetime(2023, 1, 1, 12, 0, 0, tzinfo=UTC),
            updated=datetime(2023, 1, 1, 13, 0, 0, tzinfo=UTC),
        ),
        channel=get_channel(),
    )



"""Test errors."""

from http import HTTPStatus

from ytnoti.errors import HTTPError


def test_http_errors() -> None:
    """Test creating HTTPError instances."""
    error = HTTPError("test", 400)
    assert isinstance(error.status_code, HTTPStatus)

    error = HTTPError("test", HTTPStatus.BAD_REQUEST)
    assert isinstance(error.status_code, HTTPStatus)

    assert error.message in str(error)
    assert "400" in str(error)



"""Contains the tests for the class AsyncYouTubeNotifier."""


import asyncio

import pytest

from ytnoti import AsyncYouTubeNotifier

tasks = set()


@pytest.mark.asyncio
async def test_run() -> None:
    """Test run method."""
    notifier = AsyncYouTubeNotifier()

    tasks.add(asyncio.create_task(notifier.run(port=8001)))
    await asyncio.sleep(2)
    assert notifier.is_ready


@pytest.mark.asyncio
async def test_run_in_background() -> None:
    """Test run_in_background method."""
    notifier = AsyncYouTubeNotifier()

    async with notifier.run_in_background(port=8002) as task:
        tasks.add(task)
        await asyncio.sleep(2)
        assert notifier.is_ready



"""Contains the dataclasses for the video model."""

__all__ = ["Channel", "Timestamp", "Video"]


from dataclasses import dataclass
from datetime import datetime


@dataclass
class Channel:
    """Represents a YouTube channel."""

    id: str
    """The unique ID of the channel"""

    name: str
    """The name of the channel"""

    url: str
    """The URL of the channel"""


@dataclass
class Timestamp:
    """Represents the timestamps of a video."""

    published: datetime
    """The published time of the video"""

    updated: datetime
    """The updated time of the video"""


@dataclass
class Video:
    """Represents a YouTube video."""

    id: str
    """The unique ID of the video"""

    title: str
    """The title of the video"""

    url: str
    """The URL of the video"""

    timestamp: Timestamp
    """The timestamps of the video"""

    channel: Channel
    """The channel of the video"""



"""Contains the YouTubePushNotifier class which is used to subscribe to
YouTube channels and receive push notifications when new videos are uploaded or old
videos are updated.
"""

__all__ = [
    "AsyncYouTubeNotifier",
    "Channel",
    "NotificationKind",
    "NotificationListener",
    "Timestamp",
    "Video",
    "YouTubeNotifier",
    "YouTubeNotifierConfig",
]

import asyncio
import hmac
import logging
import secrets
import signal
import string
import time
import warnings
from asyncio import Task
from collections.abc import AsyncIterator, Callable, Coroutine, Iterable, Iterator
from contextlib import asynccontextmanager, contextmanager, suppress
from datetime import datetime
from http import HTTPStatus
from pyexpat import ExpatError
from threading import Thread
from typing import Any, Literal, Self
from urllib.parse import urlparse

import xmltodict
from fastapi import APIRouter, FastAPI, Request, Response
from fastapi.routing import APIRoute
from httpx import AsyncClient
from pyngrok import ngrok
from pyngrok.exception import PyngrokNgrokURLError
from starlette.routing import Route
from uvicorn import Config, Server

from ytnoti.enums import NotificationKind
from ytnoti.errors import HTTPError
from ytnoti.models import YouTubeNotifierConfig
from ytnoti.models.history import InMemoryVideoHistory, VideoHistory
from ytnoti.models.video import Channel, Timestamp, Video
from ytnoti.types import NotificationListener, T


class AsyncYouTubeNotifier:
    """A class that encapsulates the functionality for subscribing to YouTube
    channels and receiving push notifications.
    """

    _ALL_LISTENER_KEY = "_all"

    def __init__(
        self,
        *,
        callback_url: str | None = None,
        password: str | None = None,
        video_history: VideoHistory | None = None,
    ) -> None:
        """Set up the YouTubeNotifier instance.

        :param callback_url: The URL to receive push notifications.
            If not provided, ngrok will be used to create a temporary URL.
        :param password: The password to use for verifying push notifications.
            If not provided, a random password will be generated.
        :param video_history: The video history to use to prevent duplicate
            notifications.
            If not provided,
            a new instance of InMemoryVideoHistory will be created and used.
        """
        self._logger = logging.getLogger(self.__class__.__name__)
        self._config = YouTubeNotifierConfig(
            callback_url,
            "",
            -1,
            FastAPI(),
            callback_url is None,
            password
            or str("".join(secrets.choice(string.ascii_letters) for _ in range(20))),
        )
        self._listeners: dict[
            NotificationKind, dict[str, list[NotificationListener]]
        ] = {kind: {} for kind in NotificationKind}
        self._subscribed_ids: set[str] = set()
        self._video_history = video_history or InMemoryVideoHistory()
        self._server: Server | None = None
        self._lock = asyncio.Lock()

    @property
    def callback_url(self) -> str | None:
        """Get the callback URL. If the callback URL was not provided when creating
        the instance, it will become available after the notifier is started.

        :return: The callback URL.
        """
        return self._config.callback_url

    @property
    def is_ready(self) -> bool:
        """Check if the notifier is ready to receive push notifications.

        :return: True if the notifier is ready, False otherwise.
        """
        return self._server is not None

    def listener(  # pragma: no cover
        self, *, kind: NotificationKind, channel_ids: str | Iterable[str] | None = None
    ) -> Callable[[NotificationListener], NotificationListener]:
        """Decorate the function to add a listener for push notifications.

        .. deprecated:: 2.1.0
            This method has been deprecated in favor of the more specific decorators:
            :meth:`upload`, :meth:`edit`, and :meth:`any`.
            It will be removed in version 3.0.0.

        :param kind: The kind of notification to listen for.
        :param channel_ids: The channel ID(s) to listen for.
            If not provided, the listener will be called for all channels.
        :return: The decorator function.
        :raises ValueError: If the channel ID is '_all'.
        """
        warnings.warn(
            "listener() is deprecated since version 2.1.0 and will be removed "
            "in version 3.0.0. Use upload(), edit(), or any() instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self._listener(kind=kind, channel_ids=channel_ids)

    def _listener(
        self, *, kind: NotificationKind, channel_ids: str | Iterable[str] | None = None
    ) -> Callable[[NotificationListener], NotificationListener]:
        """Decorate the function to add a listener for push notifications.

        .. deprecated:: 2.1.0
            This method has been deprecated in favor of the more specific decorators:
            :meth:`upload`, :meth:`edit`, and :meth:`any`.
            It will be removed in version 3.0.0.

        :param kind: The kind of notification to listen for.
        :param channel_ids: The channel ID(s) to listen for.
            If not provided, the listener will be called for all channels.
        :return: The decorator function.
        :raises ValueError: If the channel ID is '_all'.
        """

        def decorator(func: NotificationListener) -> NotificationListener:
            self._add_listener(func, kind, channel_ids)

            return func

        return decorator

    def any(
        self, *, channel_ids: str | Iterable[str] | None = None
    ) -> Callable[[NotificationListener], NotificationListener]:
        """Decorate the function to add a listener for any kind of push notification.
        Alias for @listener(kind=NotificationKind.ANY).

        :param channel_ids: The channel ID(s) to listen for.
            If not provided, the listener will be called for all channels.
        :return: The decorator function.
        """
        return self._listener(kind=NotificationKind.ANY, channel_ids=channel_ids)

    def upload(
        self, *, channel_ids: str | Iterable[str] | None = None
    ) -> Callable[[NotificationListener], NotificationListener]:
        """Decorate the function to add a listener for when a video is uploaded.
        Alies for @listener(kind=NotificationKind.UPLOAD).

        :param channel_ids: The channel ID(s) to listen for.
            If not provided, the listener will be called for all channels.
        :return: The decorator function.
        """
        return self._listener(kind=NotificationKind.UPLOAD, channel_ids=channel_ids)

    def edit(
        self, *, channel_ids: str | Iterable[str] | None = None
    ) -> Callable[[NotificationListener], NotificationListener]:
        """Decorate the function to add a listener for when a video is edited.
        Alies for @listener(kind=NotificationKind.EDIT).

        :param channel_ids: The channel ID(s) to listen for.
            If not provided, the listener will be called for all channels.
        :return: The decorator function.
        """
        return self._listener(kind=NotificationKind.EDIT, channel_ids=channel_ids)

    def add_listener(
        self,
        func: NotificationListener,
        kind: NotificationKind,
        channel_ids: str | Iterable[str] | None = None,
    ) -> Self:  # pragma: no cover
        """Add a listener for push notifications.

        .. deprecated:: 2.1.0
            This method has been deprecated in favor of the more specific decorators:
            :meth:`add_upload_listener`, :meth:`add_edit_listener`,
            and :meth:`add_any_listener`.
            It will be removed in version 3.0.0.

        :param func: The listener function to add.
        :param kind: The kind of notification to listen for.
        :param channel_ids: The channel ID(s) to listen for.
            If not provided, the listener will be called for all channels.
        :return: The YouTubeNotifier instance to allow for method chaining.
        :raises ValueError: If the channel ID is '_all'.
        """
        warnings.warn(
            "add_listener() is deprecated since version 2.1.0 and will be removed "
            "in version 3.0.0. Use add_upload_listener(), add_edit_listener(), "
            "or add_any_listener() instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self._add_listener(func, kind, channel_ids)

    def _add_listener(
        self,
        func: NotificationListener,
        kind: NotificationKind,
        channel_ids: str | Iterable[str] | None = None,
    ) -> Self:
        """Add a listener for push notifications.

        .. deprecated:: 2.1.0
            This method has been deprecated in favor of the more specific decorators:
            :meth:`add_upload_listener`, :meth:`add_edit_listener`,
            and :meth:`add_any_listener`.
            It will be removed in version 3.0.0.

        :param func: The listener function to add.
        :param kind: The kind of notification to listen for.
        :param channel_ids: The channel ID(s) to listen for.
            If not provided, the listener will be called for all channels.
        :return: The YouTubeNotifier instance to allow for method chaining.
        :raises ValueError: If the channel ID is '_all'.
        """
        if channel_ids is None:
            self._get_listeners(kind, None).append(func)
            self._logger.debug(
                "Added %s listener (%s) for all channels", kind.name, func.__name__
            )
            return self

        if isinstance(channel_ids, str):
            channel_ids = [channel_ids]

        for channel_id in channel_ids:
            if channel_id == self._ALL_LISTENER_KEY:
                message = f"Channel ID cannot be '{self._ALL_LISTENER_KEY}'"
                raise ValueError(message)

            self._get_listeners(kind, channel_id).append(func)
            self._logger.debug(
                "Added %s listener (%s) for channel: %s",
                kind.name,
                func.__name__,
                channel_id,
            )

        return self

    def add_any_listener(
        self, func: NotificationListener, channel_ids: str | Iterable[str] | None = None
    ) -> Self:
        """Add a listener for any kind of push notification.
        Alias for add_listener(func, NotificationKind.ANY, channel_ids).

        :param func: The listener function to add.
        :param channel_ids: The channel ID(s) to listen for.
            If not provided, the listener will be called for all channels.
        :return: The YouTubeNotifier instance to allow for method chaining.
        """
        return self._add_listener(func, NotificationKind.ANY, channel_ids)

    def add_upload_listener(
        self, func: NotificationListener, channel_ids: str | Iterable[str] | None = None
    ) -> Self:
        """Add a listener for when a video is uploaded.
        Alias for add_listener(func, NotificationKind.UPLOAD, channel_ids).

        :param func: The listener function to add.
        :param channel_ids: The channel ID(s) to listen for.
            If not provided, the listener will be called for all channels.
        :return: The YouTubeNotifier instance to allow for method chaining.
        """
        return self._add_listener(func, NotificationKind.UPLOAD, channel_ids)

    def add_edit_listener(
        self, func: NotificationListener, channel_ids: str | Iterable[str] | None = None
    ) -> Self:
        """Add a listener for when a video is edited.
        Alias for add_listener(func, NotificationKind.EDIT, channel_ids).

        :param func: The listener function to add.
        :param channel_ids: The channel ID(s) to listen for.
            If not provided, the listener will be called for all channels.
        :return: The YouTubeNotifier instance to allow for method chaining.
        """
        return self._add_listener(func, NotificationKind.EDIT, channel_ids)

    def _get_listeners(
        self, kind: NotificationKind, channel_id: str | None
    ) -> list[NotificationListener]:
        """Get the listeners for the given kind and channel ID.

        :param kind: The kind of notification.
        :param channel_id: The channel ID to get the listeners for.
            If not provided, the listeners for all channels
        :return: The listeners for the given kind and channel ID.
        """
        key = channel_id or self._ALL_LISTENER_KEY
        listeners = self._listeners[kind].get(key, None)
        if listeners is None:
            listeners = []
            self._listeners[kind][key] = listeners

        return listeners

    def _get_router(self) -> APIRouter:
        """Get the FastAPI router for the notifier.

        :return: The FastAPI router.
        """
        router = APIRouter()
        endpoint = urlparse(self._config.callback_url).path or "/"
        router.add_api_route(endpoint, self._get, methods=["HEAD", "GET"])
        router.add_api_route(endpoint, self._post, methods=["POST"])

        return router

    def _get_server_config(self, **configs: Any) -> Config:  # noqa: ANN401
        """Get the server configuration.

        :param configs: Additional arguments to pass to the server configuration.
        :return: The server configuration.
        """
        return Config(self._config.app, self._config.host, self._config.port, **configs)

    def _get_server(
        self,
        *,
        host: str,
        port: int,
        app: FastAPI | None = None,
        **configs: object,
    ) -> Server:
        """Create a server instance to receive push notifications.

        :param host: The host to run the server on.
        :param port: The port to run the server on.
        :param app: The FastAPI app to use. If not provided, a new app will be created.
        :param log_level: The log level to use for the uvicorn server.
        :param configs: Additional arguments to pass to the server configuration.
        :return: The server instance.
        :raises ValueError: If the given app instance has a route that conflicts with
            the notifier's routes.
        """
        self._config.host = host
        self._config.port = port
        self._config.app = app or self._config.app

        if self._config.using_ngrok:
            self._config.callback_url = ngrok.connect(str(port)).public_url

        self._logger.info("Callback URL: %s", self._config.callback_url)

        endpoint = urlparse(self._config.callback_url).path or "/"

        if any(
            isinstance(route, APIRoute | Route) and route.path == endpoint
            for route in self._config.app.routes
        ):
            raise ValueError(
                f"Endpoint {endpoint} is reserved for {__package__} "
                f"so it cannot be used by the app"
            )

        self._config.app.include_router(self._get_router())

        async def on_ready() -> None:
            while True:
                await asyncio.sleep(0.1)

                try:
                    async with AsyncClient() as client:
                        response = await client.head(
                            self._config.callback_url, params={"hub.challenge": "1"}
                        )
                        if response.status_code == HTTPStatus.OK:
                            break
                except ConnectionError:  # pragma: no cover
                    continue

            self._server = server

            await self._register(self._subscribed_ids)

        async def task() -> None:  # pragma: no cover
            try:
                await self._register(self._subscribed_ids)
            except RuntimeError:
                self._logger.exception("")

        self._config.app.add_event_handler(
            "startup", lambda: asyncio.create_task(on_ready())
        )
        self._config.app.add_event_handler(
            "startup",
            lambda: asyncio.create_task(self._repeat_task(task, 60 * 60 * 24)),
        )

        server = Server(self._get_server_config(**configs))
        return server

    async def _repeat_task(
        self,
        task: Callable[[], Coroutine[Any, Any, Any]],
        interval: float,
        predicate: Callable[[], bool] | None = None,
    ) -> None:
        """Repeatedly run a task.

        :param task: The coroutine to repeat
        :param interval: The interval in seconds to repeat the task
        :param predicate: An optional predicate function
            that returns True to continue
        """
        while not predicate or predicate():
            await asyncio.sleep(interval)
            try:
                await task()
            except Exception:
                self._logger.exception("Failed to repeat task")

    async def serve(
        self,
        *,
        host: str = "0.0.0.0",  # noqa: S104
        port: int = 8000,
        log_level: int = logging.WARNING,
        app: FastAPI | None = None,
        **configs: object,
    ) -> None:  # pragma: no cover
        """Start the FastAPI server to receive push notifications in an existing event
            loop and wait until the server stops.

        .. deprecated:: 2.1.0
        Use :meth:`run` instead.

        :param host: The host to run the FastAPI server on.
        :param port: The port to run the FastAPI server on.
        :param log_level: The log level to use for the uvicorn server.
        :param app: The FastAPI app instance to use. If not provided, a new instance
            will be created.
        :param configs: Additional arguments to pass to the Config class of uvicorn.
        :raises ValueError: If the given app instance has a route that conflicts with
            the notifier's routes.
        """
        warnings.warn(
            "serve() is deprecated and will be removed in "
            "version 3.0.0. Use run() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        await self.run(host=host, port=port, log_level=log_level, app=app, **configs)

    async def run(
        self,
        *,
        host: str = "0.0.0.0",  # noqa: S104
        port: int = 8000,
        log_level: int = logging.WARNING,
        app: FastAPI | None = None,
        **configs: object,
    ) -> None:
        """Start the FastAPI server to receive push notifications in an existing event
            loop and wait until the server stops.

        :param host: The host to run the FastAPI server on.
        :param port: The port to run the FastAPI server on.
        :param log_level: The log level to use for the uvicorn server.
        :param app: The FastAPI app instance to use. If not provided, a new instance
            will be created.
        :param configs: Additional arguments to pass to the Config class of uvicorn.
        :raises ValueError: If the given app instance has a route that conflicts with
            the notifier's routes.
        """
        server = self._get_server(
            host=host, port=port, app=app, log_level=log_level, **configs
        )

        old_signal_handler = signal.getsignal(signal.SIGINT)

        async def signal_handler() -> None:  # pragma: no cover
            await self._on_exit()

            signal.signal(signal.SIGINT, old_signal_handler)

        signal.signal(
            signal.SIGINT, lambda _sig, _frame: asyncio.create_task(signal_handler())
        )

        try:
            await server.serve()
        except KeyboardInterrupt:  # pragma: no cover
            await self._on_exit()

    @asynccontextmanager
    async def run_in_background(
        self,
        *,
        host: str = "0.0.0.0",  # noqa: S104
        port: int = 8000,
        app: FastAPI | None = None,
        log_level: int = logging.WARNING,
        **configs: object,
    ) -> AsyncIterator[Task]:
        """Run the FastAPI server in an existing event loop and return immediately.

        :param host: The host IP address to bind the server.
        :param port: The port number to bind the server.
        :param app: The FastAPI application to use for serving the server.
        :param log_level: The log level to use for the server.
        :param configs: Additional configurations to pass to the server.
        """
        task = asyncio.create_task(
            self.run(host=host, port=port, app=app, log_level=log_level, **configs)
        )
        try:
            while not self.is_ready:  # noqa: ASYNC110
                await asyncio.sleep(0.1)
            yield task
        finally:
            await self._on_exit()
            await task

    @staticmethod
    async def _verify_channel_ids(channel_ids: Iterable[str]) -> None:
        """Verify if the given channel IDs are valid.

        :param channel_ids: The channel IDs
        :raises ValueError: If the channel ID is invalid.
        """
        async with AsyncClient() as client:
            for channel_id in channel_ids:
                response = await client.head(
                    f"https://www.youtube.com/feeds/videos.xml?channel_id={channel_id}"
                )
                if response.status_code != HTTPStatus.OK:
                    raise ValueError(f"Invalid channel ID: {channel_id}")

    async def subscribe(self, channel_ids: str | Iterable[str]) -> Self:
        """Subscribe to YouTube channels to receive push notifications.
        This is lazy and will subscribe when the notifier is ready.
        If the notifier is already ready, it will subscribe immediately.

        :param channel_ids: The channel ID(s) to subscribe to.
        :return: The current instance for method chaining.
        :raises ValueError: If the channel ID is invalid.
        :raises HTTPError: If failed to verify the channel ID or failed to subscribe
            due to an HTTP error.
        """
        if isinstance(channel_ids, str):
            channel_ids = [channel_ids]

        await self._verify_channel_ids(channel_ids)

        if not self.is_ready:
            self._subscribed_ids.update(channel_ids)
            return self

        not_subscribed = set(channel_ids).difference(self._subscribed_ids)
        await self._register(not_subscribed)

        self._subscribed_ids.update(not_subscribed)

        return self

    async def unsubscribe(self, channel_ids: str | Iterable[str]) -> Self:
        """Unsubscribe from YouTube channels to stop receiving push notifications.

        :param channel_ids: The channel ID(s) to unsubscribe from.
        :return: The current instance for method chaining.
        :raises ValueError: If the channel_ids includes ids that are not subscribed
        """
        if isinstance(channel_ids, str):
            channel_ids = [channel_ids]

        if not self._subscribed_ids.issuperset(channel_ids):
            raise ValueError(
                f"No such subscribed channel IDs: "
                f"{self._subscribed_ids.difference(channel_ids)}"
            )

        unsubscribe_ids = self._subscribed_ids.intersection(channel_ids)

        no_server = self.is_ready

        if no_server:
            self.run_in_background(host=self._config.host, port=self._config.port)

        await self._register(unsubscribe_ids, mode="unsubscribe")

        if no_server:
            self.stop()

        self._subscribed_ids.difference_update(unsubscribe_ids)

        return self

    async def _register(
        self,
        channel_ids: Iterable[str],
        *,
        mode: Literal["subscribe", "unsubscribe"] = "subscribe",
    ) -> None:
        """Subscribe or unsubscribe to YouTube channels to receive push notifications.

        :param channel_ids: The channel ID(s) to subscribe or unsubscribe to.
        :param mode: The mode to use. Either 'subscribe' or 'unsubscribe'.
        :raises ValueError: If an invalid channel ID is provided.
        :raises ConnectionError: If this method is called while the server is not
            listening.
        :raises HTTPError: If failed to subscribe or unsubscribe due to an HTTP error.
        """
        for channel_id in channel_ids:
            async with AsyncClient() as client:
                self._logger.debug(
                    "Sending %s request for channel: %s", mode, channel_id
                )

                response = await client.post(
                    "https://pubsubhubbub.appspot.com",
                    data={
                        "hub.mode": mode,
                        "hub.topic": f"https://www.youtube.com/xml/feeds/videos.xml?channel_id={channel_id}",
                        "hub.callback": self._config.callback_url,
                        "hub.verify": "sync",
                        "hub.secret": self._config.password,
                        "hub.lease_seconds": "",
                        "hub.verify_token": "",
                    },
                    headers={"Content-type": "application/x-www-form-urlencoded"},
                )

            if response.status_code == HTTPStatus.CONFLICT:  # pragma: no cover
                if not self.is_ready:
                    raise ConnectionError(
                        f"Cannot {mode} while the server is not ready"
                    )

                raise HTTPError(
                    f"Failed to {mode} channel: {channel_id}. "
                    f"The reason might be because {self._config.callback_url} is "
                    f"inaccessible from the public internet",
                    response.status_code,
                )

            if response.status_code != HTTPStatus.NO_CONTENT:  # pragma: no cover
                raise HTTPError(
                    f"Failed to {mode} channel: {channel_id}", response.status_code
                )

            self._logger.info("Successfully %sd channel: %s", mode, channel_id)

    def stop(self) -> None:
        """Gracefully stop the FastAPI server and the notifier.
        If the notifier is not running, this method will do nothing.
        """
        if not self.is_ready:
            return

        self._server.should_exit = True
        self._server = None

        if self._config.using_ngrok:
            with suppress(PyngrokNgrokURLError):
                ngrok.disconnect(self._config.callback_url)

    async def _on_exit(self) -> None:
        """Perform a task after the notifier is stopped."""
        if not self._config.using_ngrok:
            await self.unsubscribe(self._subscribed_ids)

        self.stop()

    @staticmethod
    async def _get(request: Request) -> Response:
        """Handle a challenge from the Google pubsubhubbub server."""
        challenge = request.query_params.get("hub.challenge")
        if challenge is None:
            return Response(status_code=HTTPStatus.BAD_REQUEST)

        return Response(challenge)

    async def _post(self, request: Request) -> Response:
        """Handle push notifications from the Google pubsubhubbub server."""
        if not await self._is_authorized(request):
            return Response(status_code=HTTPStatus.UNAUTHORIZED)

        body = await request.body()

        try:
            body = xmltodict.parse(body)
        except ExpatError:
            self._logger.debug("Received invalid request body: %s", body)
            return Response(status_code=HTTPStatus.BAD_REQUEST)

        self._logger.debug("Received push notification: %s", body)

        try:
            if "at:deleted-entry" in body["feed"]:
                self._logger.debug("Ignoring push notification for deleted video")
                return Response(status_code=HTTPStatus.NO_CONTENT)

            # entry can be list of dict or just dict
            entries = (
                body["feed"]["entry"]
                if isinstance(body["feed"]["entry"], list)
                else [body["feed"]["entry"]]
            )

            for entry in entries:
                channel = Channel(
                    id=entry["yt:channelId"],
                    name=entry["author"]["name"],
                    url=entry["author"]["uri"],
                )

                timestamp = Timestamp(
                    published=self._parse_timestamp(entry["published"]),
                    updated=self._parse_timestamp(entry["updated"]),
                )

                url = (
                    entry["link"][0]["@href"]
                    if isinstance(entry["link"], list)
                    else entry["link"]["@href"]
                )

                video_id = entry["yt:videoId"]

                video = Video(
                    id=video_id,
                    title=entry["title"],
                    url=url,
                    timestamp=timestamp,
                    channel=channel,
                )

                async with self._lock:
                    if (
                        timestamp.published == timestamp.updated
                        or not await self._video_history.has(video)
                    ):
                        kind = NotificationKind.UPLOAD
                        await self._video_history.add(video)
                    else:
                        kind = NotificationKind.EDIT

                self._logger.debug("Classified video (%s) as %s", video.id, kind)

                listeners = (
                    self._get_listeners(kind, None)
                    + self._get_listeners(kind, channel.id)
                    + self._get_listeners(NotificationKind.ANY, None)
                    + self._get_listeners(NotificationKind.ANY, channel.id)
                )

                for func in listeners:
                    await func(video)
        except (TypeError, KeyError, ValueError) as ex:
            raise RuntimeError(f"Failed to parse request body: {body}") from ex

        return Response(status_code=HTTPStatus.NO_CONTENT)

    @staticmethod
    def _parse_timestamp(timestamp: str) -> datetime:
        time, zone = timestamp.split("+", 1)

        # Remove fractional seconds if exists
        time = time.split(".", 1)[0]

        return datetime.strptime(f"{time}+{zone}", "%Y-%m-%dT%H:%M:%S%z")

    async def _is_authorized(self, request: Request) -> bool:
        if not self._config.password:
            return True

        x_hub_signature = request.headers.get("X-Hub-Signature")
        # Check if the header is missing or invalid
        if x_hub_signature is None or "=" not in x_hub_signature:
            return False

        algorithm, value = x_hub_signature.split("=")
        hash_obj = hmac.new(
            self._config.password.encode(), await request.body(), algorithm
        )
        return hmac.compare_digest(hash_obj.hexdigest(), value)


class YouTubeNotifier(AsyncYouTubeNotifier):
    """A class that encapsulates the functionality for subscribing to YouTube
    channels and receiving push notifications.
    """

    def __init__(
        self,
        *,
        callback_url: str | None = None,
        password: str | None = None,
        video_history: VideoHistory | None = None,
    ) -> None:
        """Create a new YouTubeNotifier instance.

        :param callback_url: The URL to receive push notifications.
            If not provided, ngrok will be used to create a temporary URL.
        :param password: The password to use for verifying push notifications.
            If not provided, a random password will be generated.
        :param video_history: The video history to use to prevent duplicate
            notifications.
            If not provided, a new instance of InMemoryVideoHistory will be created and
            used.
        """
        self._logger = logging.getLogger(self.__class__.__name__)
        super().__init__(
            callback_url=callback_url,
            password=password,
            video_history=video_history,
        )

    def subscribe(self, channel_ids: str | Iterable[str]) -> Self:  # noqa: D102
        return self._run_coroutine(super().subscribe(channel_ids))

    def unsubscribe(self, channel_ids: str | Iterable[str]) -> Self:  # noqa: D102
        return self._run_coroutine(super().unsubscribe(channel_ids))

    def run(
        self,
        *,
        host: str = "0.0.0.0",  # noqa: S104
        port: int = 8000,
        app: FastAPI | None = None,
        log_level: int = logging.WARNING,
        **configs: object,
    ) -> None:
        """Start the FastAPI server to receive push notifications in the
            current thread and wait until the server stops.

        :param host: The host to run the FastAPI server on.
        :param port: The port to run the FastAPI server on.
        :param log_level: The log level to use for the uvicorn server.
        :param app: The FastAPI app instance to use. If not provided, a new instance
            will be created.
        :param configs: Additional arguments to pass to the Config class of uvicorn.
        :raises ValueError: If the given app instance has a route that conflicts with
            the notifier's routes.
        """
        server = self._get_server(
            host=host, port=port, app=app, log_level=log_level, **configs
        )

        try:
            server.run()
        except KeyboardInterrupt:  # pragma: no cover
            pass
        finally:
            self._on_exit()

    @contextmanager
    def run_in_background(
        self,
        *,
        host: str = "0.0.0.0",  # noqa: S104
        port: int = 8000,
        app: FastAPI | None = None,
        log_level: int = logging.WARNING,
        **configs: object,
    ) -> Iterator[Thread]:
        """Start the FastAPI server to receive push notifications in the separate
            thread and return immediately.

        :param host: The host to run the FastAPI server on.
        :param port: The port to run the FastAPI server on.
        :param log_level: The log level to use for the uvicorn server.
        :param app: The FastAPI app instance to use. If not provided, a new instance
            will be created.
        :param configs: Additional arguments to pass to the Config class of uvicorn.
        :return: A thread that runs the FastAPI server in the background.
        :raises ValueError: If the given app instance has a route that conflicts with
            the notifier's routes.
        """
        configs["host"] = host
        configs["port"] = port
        configs["app"] = app
        configs["log_level"] = log_level

        thread = Thread(target=self.run, kwargs=configs, daemon=True)
        thread.start()
        try:
            while not self.is_ready:
                time.sleep(0.1)
            yield thread
        finally:
            self._on_exit()
            thread.join()

    def _on_exit(self) -> None:
        if not self._config.using_ngrok:
            self.unsubscribe(self._subscribed_ids)

        self.stop()

    @staticmethod
    def _run_coroutine(coro: Coroutine[Any, Any, T]) -> T:
        """Run a coroutine in the event loop.

        :param coro: The coroutine to run.
        :return: The result of the coroutine.
        """
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            return asyncio.run(coro)
        else:
            return loop.run_until_complete(coro)



"""Test class InMemoryFileHistory."""

import pytest

from tests import get_video
from ytnoti.models.history import InMemoryVideoHistory

CACHE_SIZE = 100

@pytest.fixture
def history() -> InMemoryVideoHistory:
    """Create a mock FileVideoHistory instance."""
    return InMemoryVideoHistory(cache_size=CACHE_SIZE)


def test_cache_size(history: InMemoryVideoHistory) -> None:
    """Test the cache size of the InMemoryVideoHistory class."""
    assert history.cache_size == CACHE_SIZE
    history.cache_size = 1
    assert history.cache_size == 1


@pytest.mark.asyncio
async def test_has(history: InMemoryVideoHistory) -> None:
    """Test the has method of the InMemoryVideoHistory class."""
    video = get_video()

    assert not await history.has(video)

    history._video_ids[video.id] = None

    assert await history.has(video)

@pytest.mark.asyncio
async def test_add(history: InMemoryVideoHistory) -> None:
    """Test the add method of the InMemoryVideoHistory class."""
    video = get_video()
    video.id = "-1"

    await history.add(video)
    await history.add(video)
    assert len(history._video_ids) == 1

    assert video.id in history._video_ids

    for i in range(CACHE_SIZE + 1):
        new_video = get_video()
        new_video.id = str(i)
        await history.add(new_video)

    assert len(history._video_ids) == CACHE_SIZE
    assert video.id not in history._video_ids



"""Defines Enum classes used in the package."""

__all__ = ["NotificationKind"]

from enum import Enum


class NotificationKind(Enum):
    """Enum for the kind of notification."""

    UPLOAD = 0
    """Upload notification"""

    EDIT = 1
    """Edit notification"""

    ANY = 2
    """Any notification"""



"""Contains the tests for the class YouTubeNotifier."""

from collections.abc import Iterator
from datetime import UTC, datetime
from http import HTTPStatus
from threading import Thread

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient

from tests import CALLBACK_URL
from ytnoti import Video, YouTubeNotifier

channel_ids = [
    "UCPF-oYb2-xN5FbCXy0167Gg",
    "UCuFFtHWoLl5fauMMD5Ww2jA",
    "UCupvZG-5ko_eiXAupbDfxWw",
]

# ruff: noqa: E501 ERA001

xmls = [
    """
    <feed xmlns:yt="http://www.youtube.com/xml/schemas/2015" xmlns="http://www.w3.org/2005/Atom">
      <link rel="hub" href="https://pubsubhubbub.appspot.com"/>
      <link rel="self" href="https://www.youtube.com/xml/feeds/videos.xml?channel_id=CHANNEL_ID"/>
      <title>YouTube video feed</title>
      <updated>2015-04-01T19:05:24.552394234+00:00</updated>
      <entry>
        <id>yt:video:VIDEO_ID</id>
        <yt:videoId>VIDEO_ID</yt:videoId>
        <yt:channelId>CHANNEL_ID</yt:channelId>
        <title>Video title</title>
        <link rel="alternate" href="http://www.youtube.com/watch?v=VIDEO_ID"/>
        <author>
         <name>Channel title</name>
         <uri>http://www.youtube.com/channel/CHANNEL_ID</uri>
        </author>
        <published>2015-03-06T21:40:57+00:00</published>
        <updated>2015-03-09T19:05:24.552394234+00:00</updated>
      </entry>
    </feed>
    """,
    """
    <feed xmlns:yt="http://www.youtube.com/xml/schemas/2015" xmlns="http://www.w3.org/2005/Atom">
      <link rel="hub" href="https://pubsubhubbub.appspot.com"/>
      <link rel="self" href="https://www.youtube.com/xml/feeds/videos.xml?channel_id=CHANNEL_ID"/>
      <title>YouTube video feed</title>
      <updated>2015-04-01T19:05:24.552394234+00:00</updated>
      <entry>
        <id>yt:video:VIDEO_ID</id>
        <yt:videoId>VIDEO_ID</yt:videoId>
        <yt:channelId>CHANNEL_ID</yt:channelId>
        <title>Video title</title>
        <link rel="alternate" href="http://www.youtube.com/watch?v=VIDEO_ID"/>
        <link rel="alternate" href="http://www.youtube.com/watch?v=VIDEO_ID"/>
        <link rel="alternate" href="http://www.youtube.com/watch?v=VIDEO_ID"/>
        <link rel="alternate" href="http://www.youtube.com/watch?v=VIDEO_ID"/>
        <author>
         <name>Channel title</name>
         <uri>http://www.youtube.com/channel/CHANNEL_ID</uri>
        </author>
        <published>2015-03-06T21:40:57+00:00</published>
        <updated>2015-03-09T19:05:24.552394234+00:00</updated>
      </entry>
    </feed>
    """,
    """
    <feed xmlns:yt="http://www.youtube.com/xml/schemas/2015" xmlns="http://www.w3.org/2005/Atom">
      <link rel="hub" href="https://pubsubhubbub.appspot.com"/>
      <link rel="self" href="https://www.youtube.com/xml/feeds/videos.xml?channel_id=CHANNEL_ID"/>
      <title>YouTube video feed</title>
      <updated>2015-04-01T19:05:24.552394234+00:00</updated>
      <entry>
        <id>yt:video:VIDEO_ID</id>
        <yt:videoId>VIDEO_ID</yt:videoId>
        <yt:channelId>CHANNEL_ID</yt:channelId>
        <title>Video title</title>
        <link rel="alternate" href="http://www.youtube.com/watch?v=VIDEO_ID"/>
        <author>
         <name>Channel title</name>
         <uri>http://www.youtube.com/channel/CHANNEL_ID</uri>
        </author>
        <published>2015-03-06T21:40:57+00:00</published>
        <updated>2015-03-09T19:05:24.552394234+00:00</updated>
      </entry>
      <entry>
        <id>yt:video:VIDEO_ID</id>
        <yt:videoId>VIDEO_ID</yt:videoId>
        <yt:channelId>CHANNEL_ID</yt:channelId>
        <title>Video title</title>
        <link rel="alternate" href="http://www.youtube.com/watch?v=VIDEO_ID"/>
        <author>
         <name>Channel title</name>
         <uri>http://www.youtube.com/channel/CHANNEL_ID</uri>
        </author>
        <published>2015-03-06T21:40:57+00:00</published>
        <updated>2015-03-09T19:05:24.552394234+00:00</updated>
      </entry>
    </feed>
    """,
    """
    <feed xmlns:at="http://purl.org/atompub/tombstones/1.0" xmlns="http://www.w3.org/2005/Atom">
        <at:deleted-entry ref="yt:video:VIDEO_ID" when="2024-09-09T22:34:19.642702+00:00">
            <link href="https://www.youtube.com/watch?v=VIDEO_ID" />
            <at:by>
                <name>Channel title</name>
                <uri>https://www.youtube.com/channel/CHANNEL_ID</uri>
            </at:by>
        </at:deleted-entry>
    </feed>
    """,
]

# ruff: enable


@pytest.fixture(scope="session")
def notifier() -> Iterator[YouTubeNotifier]:
    """Setup/Teardown code that runs before and after the tests in this package."""
    noti = YouTubeNotifier()
    noti._config.password = ""

    with noti.run_in_background():
        yield noti


def test_callback_url(notifier: YouTubeNotifier) -> None:
    """Test the callback URL configuration."""
    assert notifier.callback_url is not None


def test_subscribe(notifier: YouTubeNotifier) -> None:
    """Test the subscribe method of the YouTubeNotifier class."""
    notifier.subscribe(channel_ids)

    with pytest.raises(ValueError):
        notifier.subscribe("Invalid")

    notifier = YouTubeNotifier()
    notifier.subscribe(channel_ids)
    assert len(notifier._subscribed_ids) == len(channel_ids)


def test_unsubscribe(notifier: YouTubeNotifier) -> None:
    """Test the unsubscribe method of the YouTubeNotifier class."""
    notifier.subscribe(channel_ids)
    notifier.unsubscribe(channel_ids)

    with pytest.raises(ValueError):
        notifier.unsubscribe("Invalid")

    with pytest.raises(ValueError):
        notifier.unsubscribe(channel_ids)

    assert len(notifier._subscribed_ids) == 0


def test_listener(notifier: YouTubeNotifier) -> None:
    """Test the upload decorator of the YouTubeNotifier class."""
    client = TestClient(notifier._config.app)
    content = xmls[0]
    any_called = False
    upload_called = False

    @notifier.any()
    async def listener(_video: Video) -> None:
        nonlocal any_called
        any_called = True

    @notifier.upload()
    async def listener(_video: Video) -> None:
        nonlocal upload_called
        upload_called = True

    client.post(CALLBACK_URL, content=content)

    assert any_called
    assert upload_called

    upload_called = False
    edit_called = False

    @notifier.edit()
    async def listener(_video: Video) -> None:
        nonlocal edit_called
        edit_called = True

    client.post(CALLBACK_URL, content=content)

    assert any_called
    assert edit_called
    assert not upload_called


def test_listener_channel_id(notifier: YouTubeNotifier) -> None:
    """Test the listener decorator with channel ID."""
    client = TestClient(notifier._config.app)
    content = xmls[0]

    called = 0

    @notifier.any(channel_ids="CHANNEL_ID")
    async def listener(_video: Video) -> None:
        nonlocal called
        called += 1

    @notifier.any(channel_ids=["CHANNEL_ID", "invalid"])
    async def listener(_video: Video) -> None:
        nonlocal called
        called += 1

    @notifier.any(channel_ids=["invalid"])
    async def listener(_video: Video) -> None:
        nonlocal called
        called += 1

    client.post(CALLBACK_URL, content=content)
    assert called == 2


def test_add_listener(notifier: YouTubeNotifier) -> None:
    """Test adding listeners to the YouTubeNotifier class."""

    async def listener1(_video: Video) -> None:
        pass

    async def listener2(_video: Video) -> None:
        pass

    async def listener3(_video: Video) -> None:
        pass

    notifier.add_any_listener(listener1)
    notifier.add_upload_listener(listener2)
    notifier.add_edit_listener(listener3)

    assert len(notifier._listeners) == 3

    with pytest.raises(ValueError):
        notifier.add_any_listener(listener1, channel_ids=notifier._ALL_LISTENER_KEY)


def test_get_server(notifier: YouTubeNotifier) -> None:
    """Test getting the server of the YouTubeNotifier class."""
    host = "0.0.0.0"  # noqa: S104
    port = 8000
    app = FastAPI()

    using_ngrok = notifier._config.using_ngrok
    notifier._config.using_ngrok = False
    notifier._get_server(host=host, port=port, app=app)

    app.get("/")(lambda: None)

    with pytest.raises(ValueError):
        notifier._get_server(host=host, port=port, app=app)

    notifier._config.using_ngrok = using_ngrok


@pytest.mark.asyncio
async def test_verify_channel_ids() -> None:
    """Test verifying channel IDs."""
    notifier = YouTubeNotifier()
    await notifier._verify_channel_ids(channel_ids)

    with pytest.raises(ValueError):
        await notifier._verify_channel_ids(["Invalid"])


def test_get(notifier: YouTubeNotifier) -> None:
    """Test the get method of the YouTubeNotifier class."""
    client = TestClient(notifier._config.app)

    response = client.get(CALLBACK_URL)
    assert response.status_code == HTTPStatus.BAD_REQUEST

    response = client.get(CALLBACK_URL, params={"hub.challenge": 1})
    assert response.status_code == HTTPStatus.OK


def test_parse_timestamp(notifier: YouTubeNotifier) -> None:
    """Test parsing timestamp."""
    timestamp = "2015-04-01T19:05:24.552394234+00:00"
    parsed_timestamp = notifier._parse_timestamp(timestamp)
    assert parsed_timestamp == datetime(2015, 4, 1, 19, 5, 24, tzinfo=UTC)


def test_post(notifier: YouTubeNotifier) -> None:
    """Test the post method of the YouTubeNotifier class."""
    client = TestClient(notifier._config.app)

    headers = {"Content-Type": "application/xml"}
    for xml in xmls:
        response = client.post(CALLBACK_URL, headers=headers, content=xml)
        assert response.status_code == HTTPStatus.NO_CONTENT

    response = client.post(CALLBACK_URL, headers=headers, content="Invalid")
    assert response.status_code == HTTPStatus.BAD_REQUEST

    with pytest.raises(RuntimeError):
        client.post(CALLBACK_URL, headers=headers, content="<feed/>")

    password = notifier._config.password
    notifier._config.password = "password"  # noqa: S105
    response = client.post(CALLBACK_URL, headers=headers, content=xmls[0])
    assert response.status_code == HTTPStatus.UNAUTHORIZED

    response = client.post(
        CALLBACK_URL,
        headers={**headers, "X-Hub-Signature": "sha256=password"},
        content=xmls[0],
    )
    assert response.status_code == HTTPStatus.UNAUTHORIZED

    notifier._config.password = password


def test_post_race_condition() -> None:
    """Test the post method of the YouTubeNotifier class with race condition."""
    notifier = YouTubeNotifier()
    client = TestClient(notifier._config.app)
    headers = {"Content-Type": "application/xml"}

    num_called = 0

    @notifier.upload()
    async def listener(_video: Video) -> None:
        nonlocal num_called
        num_called += 1

    def send_post() -> None:
        client.post(CALLBACK_URL, headers=headers, content=xmls[0])

    password = notifier._config.password
    notifier._config.password = ""

    threads = [Thread(target=send_post) for _ in range(2)]

    with notifier.run_in_background():
        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

    notifier._config.password = password

    assert num_called == 1


@pytest.mark.asyncio
async def test_repeat_task() -> None:
    """Test repeat method of the YouTubeNotifier class."""
    notifier = YouTubeNotifier()

    timeout = 3
    started = datetime.now(UTC)

    called = 0

    async def task() -> None:
        nonlocal called
        called += 1

    await notifier._repeat_task(
        task, 1, lambda: (datetime.now(UTC) - started).total_seconds() < timeout
    )

    assert timeout - 1 <= called <= timeout



"""Test the FileVideoHistory class."""
import shutil
from pathlib import Path

import pytest

from tests import get_video
from ytnoti.models.history import FileVideoHistory

NUM_VIDEOS = 100

@pytest.fixture
def history() -> FileVideoHistory:
    """Create a FileVideoHistory instance."""
    path = Path("./history")
    path.mkdir(parents=True, exist_ok=True)

    yield FileVideoHistory(dir_path=path, num_videos=NUM_VIDEOS)

    shutil.rmtree(path)


@pytest.mark.asyncio
async def test_has(history: FileVideoHistory) -> None:
    """Test the has method of the FileVideoHistory class."""
    video = get_video()
    path = history._get_path(video.channel)

    assert not await history.has(video)

    with path.open("w") as file:
        file.write(video.id)

    assert await history.has(video)

    new_video = get_video()
    new_video.id = "-1"
    assert not await history.has(new_video)

@pytest.mark.asyncio
async def test_truncate(history: FileVideoHistory) -> None:
    """Test the truncate method of the FileVideoHistory class."""
    video = get_video()
    path = history._get_path(video.channel)

    # it should be no-op if the file does not exist
    await history._truncate(video.channel)

    with path.open("w") as file:
        for _ in range(NUM_VIDEOS + 5):
            file.write(video.id + "\n")

    await history._truncate(video.channel)

    with path.open("r") as file:
        lines = file.readlines()
        assert len(lines) == NUM_VIDEOS


@pytest.mark.asyncio
async def test_add(history: FileVideoHistory) -> None:
    """Test the add method of the FileVideoHistory class."""
    video = get_video()
    path = history._get_path(video.channel)

    await history.add(video)

    with path.open("r") as file:
        assert file.read().strip() == video.id

    for _ in range(NUM_VIDEOS + 5):
        await history.add(video)

    with path.open("r") as file:
        lines = file.readlines()
        assert len(lines) == NUM_VIDEOS



"""Contains custom exceptions for the ytnoti package."""

from http import HTTPStatus


class HTTPError(Exception):
    """Exception raised when an HTTP error occurs."""

    def __init__(self, message: str, status_code: int | HTTPStatus) -> None:
        """Initialize the HTTPError object.

        :param message: The error message
        :param status_code: The status code of the error
        """
        self.status_code = (
            status_code
            if isinstance(status_code, HTTPStatus)
            else HTTPStatus(status_code)
        )
        self.message = message

    def __str__(self) -> str:
        """Return a string representation of the HTTPError object."""
        return f"Status code: {self.status_code}: {self.message}"



"""Contains type hints for the library."""

from collections.abc import Callable, Coroutine
from typing import Any, TypeVar

from ytnoti.models.video import Video

NotificationListener = Callable[[Video], Coroutine[Any, Any, Any]]

T = TypeVar("T")



"""Contains the video history model."""

import logging
from abc import ABC, abstractmethod
from asyncio import Lock
from collections import OrderedDict
from os import PathLike
from pathlib import Path

import aiofiles
from aiofiles import os, ospath

from ytnoti.models.video import Channel, Video


class VideoHistory(ABC):
    """Represents a history of videos."""

    @abstractmethod
    async def add(self, video: Video) -> None:
        """Add a video to the history.

        :param video: The video to add.
        """

    @abstractmethod
    async def has(self, video: Video) -> bool:
        """Check if a video is in the history.

        :param video: The video to check.
        :return: True if the video is in the history, False otherwise.
        """


class InMemoryVideoHistory(VideoHistory):
    """Represents an in-memory history of notifications."""

    def __init__(self, *, cache_size: int = 5000) -> None:
        """Create a new InMemoryVideoHistory instance.

        :param cache_size: The size of the cache.
            If the cache is full, the oldest video will be removed.
        """
        self._logger = logging.getLogger(self.__class__.__name__)
        self._video_ids: OrderedDict[str, None] = OrderedDict()
        self._cache_size = cache_size
        self._lock = Lock()

    @property
    def cache_size(self) -> int:
        """Get the size of the cache.

        :return: The size of the cache.
        """
        return self._cache_size

    @cache_size.setter
    def cache_size(self, value: int) -> None:
        self._logger.debug("Setting cache size to %d", value)
        self._cache_size = value

    async def add(self, video: Video) -> None:
        """Add a video to the history.

        :param video: The video to add.
        """
        async with self._lock:
            if video.id in self._video_ids:
                self._logger.debug("Skipping to add video (%s) that already "
                                   "exists in history", video.id)
                return

            if len(self._video_ids) >= self._cache_size:
                self._logger.debug("Removing oldest video from history")
                self._video_ids.popitem(last=False)

            self._logger.debug("Adding video (%s) to history", video.id)
            self._video_ids[video.id] = None

    async def has(self, video: Video) -> bool:
        """Check if a video is in the history.

        :param video: The video to check.
        """
        async with self._lock:
            return video.id in self._video_ids


class FileVideoHistory(VideoHistory):
    """Represents a file-based history of videos."""

    def __init__(self, *, dir_path: str | PathLike[str], num_videos: int = 100) -> None:
        """Create a new FileVideoHistory instance.

        :param dir_path: The path to the directory to store the history files.
        :param num_videos: The number of videos to keep in the history file.
            If the number of videos exceeds this value, the oldest videos will be
            removed.
        """
        self._logger = logging.getLogger(self.__class__.__name__)
        self._dir_path = Path(dir_path)
        self._num_videos = num_videos
        self._lock = Lock()

    def _get_path(self, channel: Channel) -> Path:
        """Get the path to the history file for a channel.

        :param channel: The channel to get the history file for.
        """
        return self._dir_path / channel.id

    async def _truncate(self, channel: Channel) -> None:
        """Truncate the history file for a channel."""
        path = self._get_path(channel)

        if not await ospath.exists(path):
            return

        async with aiofiles.open(path, encoding="utf-8") as file:
            lines = await file.readlines()

        async with aiofiles.open(path, "w", encoding="utf-8") as file:
            await file.writelines(lines[-self._num_videos :])

    async def add(self, video: Video) -> None:
        """Add a video to the history.
        :param video: The video to add.
        """
        await os.wrap(self._dir_path.mkdir)(parents=True, exist_ok=True)

        path = self._get_path(video.channel)

        async with self._lock:
            await os.makedirs(path.parent, exist_ok=True)

            async with aiofiles.open(path, "a", encoding="utf-8") as file:
                self._logger.debug("Adding video (%s) to history at %s", video.id, path)
                await file.write(f"{video.id}\n")

            await self._truncate(video.channel)

    async def has(self, video: Video) -> bool:
        """Check if a video is in the history.

        :param video: The video to check.
        """
        path = self._get_path(video.channel)

        async with self._lock:
            if not await ospath.exists(path):
                return False

            async with aiofiles.open(path, encoding="utf-8") as file:
                async for line in file:
                    if line.strip() == video.id:
                        return True

            return False



"""Setup script for ytnoti."""

from pathlib import Path

from setuptools import find_packages, setup

with Path("README.md").open() as file:
    long_description = file.read()

setup(
    name="ytnoti",
    version="2.1.3",
    packages=find_packages(),
    author="SeoulSKY",
    author_email="contact@seoulsky.dev",
    description="Easy-to-use Python library for receiving YouTube push notification "
    "for video upload and edit in real-time",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "fastapi~=0.115.0",
        "httpx~=0.28.1",
        "uvicorn~=0.34.0",
        "xmltodict~=0.14.2",
        "pyngrok~=7.2.3",
        "aiofiles~=24.1.0",
    ],
    url="https://github.com/SeoulSKY/ytnoti",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
)
